/*
 * Copyright 1991 Addison-Wesley Publishing Company
 *                (see the file "COPYRIGHTS" for details)
 */

/*****************************************************************************
*
*        motifhelp.c
*
*        Displays Info about Motif Widget Classes
*
*        Author:
*        Thomas Berlage, GMD
*        P.O. Box 1240
*        D-5205 Sankt Augustin 1
*        Phone: +49 (2241) 14-0
*        Email: berlage@gmdzi.gmd.de
*
*        Version 1.2
*        05.07.90
*
*****************************************************************************/

#include "motifhelp.h"

#include <stdio.h>
#include <strings.h>

#include <Xm/Xm.h>
#include <X11/Shell.h>
#include <Xm/MainW.h>
#include <Xm/Text.h>
#include <Xm/RowColumn.h>
#include <Xm/BulletinB.h>
#include <Xm/CascadeB.h>
#include <Xm/CascadeBG.h>
#include <Xm/DrawnB.h>
#include <Xm/Frame.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/List.h>
#include <Xm/MessageB.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/SelectioB.h>
#include <Xm/ScrolledW.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/Separator.h>
#include <Xm/SeparatoG.h>
#include <Xm/PanedW.h>

/********************** Forward Declarations ************************/

/* menu system */
Widget CreateMenus ();
void SetSectionSensitivity ();
void SetSubclassesMenu ();

/* main form */
void CreateMainForm ();

/* superclass chain */
void CreateClassChain ();
void ChangeClassChain ();
ClassDoc *ActiveDoc ();

/* resource selection list */
void CreateSelectionList ();
void SetResourceSelectionList ();
void SelectResourceEntry ();

/* parameter field */
void CreateParameterField ();
void ParamVisibility ();
Boolean ParamIsVisible ();
void SetParameterField ();

/* help text area */
void CreateHelpText ();
void SetHelpText ();

/* class selection dialog */
void CreateClassSelectionDialog ();
void PopupClassDialog ();

/* resource selection dialog */
void CreateResSelectionDialog ();
void PopupResDialog ();

/* help dialog */
void CreateHelpDialog ();
void PopupHelpDialog ();

/*************************** Global Data ******************************/

#define NUM_ARGS 10     /* max. size of argument list in this program */
#define SMALL_BUFLENGTH 130


/******************************* Main Program ****************************/

#define PROGRAM_CLASS "Mhelp"

#define MAIN_TITLE     "Motif Help Browser"

/* structure for additional application resources */

typedef struct {
    String    help_dir;
} AppData, *AppDataPtr;

static XtResource resources [] = {
    {"helpDir", "Dir", XmRString, sizeof (String),
     XtOffset (AppDataPtr, help_dir), XmRString, ""}
};

/* additional command line options */

XrmOptionDescRec options [] = {
    {"-dir", "*helpDir", XrmoptionSepArg, ""}
};

/* main program */

main (argc, argv)
    unsigned int   argc;
    char         **argv;
{
    Display       *display;
    XtAppContext   ac;
    AppData        data;
    Arg            args[NUM_ARGS];
    Widget         toplevel, bar, main_w, menu,pane;
    Widget         frame1, form;

    /* init toolkit */
    XtToolkitInitialize();
    ac = XtCreateApplicationContext();
    display = XtOpenDisplay (
                  ac,          /* application context = default */
                  NULL,        /* use DISPLAY from environment  */
                  NULL,        /* use last of argv[0] as name   */
                  PROGRAM_CLASS,
                  options,     /* add'tl cmd line options       */
                  XtNumber (options),
                  &argc,
                  argv);       /* use and delete std options    */

    if (display == NULL)       /* necessary !                   */
        XtErrorMsg (
            NErrCannotOpen, TErrMain,  /* resource name         */
            ERROR_CLASS,               /* resource class        */
            "cannot open display",     /* default message       */
            NULL, NULL);               /* parameters            */

    /* create application shell */
    XtSetArg (args[0], XmNtitle, MAIN_TITLE);
    toplevel = XtAppCreateShell (
                   NULL,               /* use same program name */
                   PROGRAM_CLASS,      /* repeat class param    */
                   applicationShellWidgetClass,
                   display,
                   args, 1);           /* argument list         */

    /* read additional resources */
    XtGetApplicationResources (toplevel, &data, resources,
                               XtNumber (resources), NULL, 0);
    SetHelpDir (data.help_dir);

    /* create main window */
    main_w = XtCreateManagedWidget ("main", xmMainWindowWidgetClass, toplevel, 
                               NULL, 0);

    /* create menus */
    bar = CreateMenus (main_w);

    /* create paned window */
    pane = XtCreateManagedWidget ("pane", xmPanedWindowWidgetClass, main_w,
                                  args, 0);

    /* create selection pane ... */
    CreateMainForm (pane);

    /* create text pane ... */
    CreateHelpText (pane);

    XmMainWindowSetAreas (main_w, bar, NULL, NULL, NULL, pane);

    /* create dialog shells */
    CreateClassSelectionDialog (bar);
    CreateResSelectionDialog (bar);
    CreateHelpDialog (bar);

    /* set initial class */
    ChangeClassChain (SearchClass (MakeXmString ("FileSelectionBox")));

    /* start loop */
    XtRealizeWidget (toplevel);
    XtAppMainLoop(ac);
}

/********************************* Menus **********************************/

/* widget ID's for all the menu buttons */

#define MAX_MENU_BUTTONS 8

static Widget buttons1[MAX_MENU_BUTTONS], 
              buttons2[MAX_MENU_BUTTONS], 
              buttons3[MAX_MENU_BUTTONS], 
              buttons4[MAX_MENU_BUTTONS],
              buttons5[MAX_MENU_BUTTONS],
              buttons6[MAX_MENU_BUTTONS],
              buttons7[MAX_MENU_BUTTONS];

/* identify button index from callback */

int ButtonIndex (buttons, call_data)
    Widget    buttons[MAX_MENU_BUTTONS];
    XmRowColumnCallbackStruct  *call_data;
{
    int       i;

    for (i=0; i<MAX_MENU_BUTTONS; i++)
        if (buttons[i] == call_data->widget) return i;
    XtErrorMsg (
        NErrInternal, TErrMenu,    /* resource name         */
        ERROR_CLASS,               /* resource class        */
        "button not found",        /* default message       */
        NULL, NULL);               /* parameters            */
    return -1;
}

/*** File Menu ***/

void HandleFileMenu (w, client_data, call_data)
    Widget w;
    caddr_t client_data;
    XmRowColumnCallbackStruct *call_data;
{
    exit (0);
}

/*** Edit Menu ***/

void HandleEditMenu (w, client_data, call_data)
    Widget w;
    caddr_t client_data;
    XmRowColumnCallbackStruct *call_data;
{
}

/*** Locate Menu ***/

void HandleLocateMenu (w, client_data, call_data)
    Widget w;
    caddr_t client_data;
    XmRowColumnCallbackStruct *call_data;
{
    switch (ButtonIndex (buttons3, call_data)) {
    case 0:
        PopupClassDialog (); 
        break;
    case 1:
        PopupResDialog();
        break;
    }
}

/*** Option Menu ***/

void HandleOptionMenu (w, client_data, call_data)
    Widget w;
    caddr_t client_data;
    XmRowColumnCallbackStruct *call_data;
{
    int    who = ButtonIndex (buttons4, call_data);

    ParamVisibility (who, ! ParamIsVisible (who));
}

/*** Info Menu ***/

void HandleInfoMenu (w, client_data, call_data)
    Widget w;
    caddr_t client_data;
    XmRowColumnCallbackStruct *call_data;
{
    Arg        args[NUM_ARGS];
    String     buffer;

    buffer = ReadSection (ActiveDoc (), ButtonIndex (buttons5, call_data) + 1);
    SetHelpText (buffer);
}

/* Adjust sensitivity of menu buttons to reflect which sections exist */

void SetSectionSensitivity (doc)
    ClassDoc  *doc;
{
    int        i, mask;

    mask = 1;
    for (i=0; i<COLOR_SEC; i++) {
        XtSetSensitive (buttons5[i], doc->sections & mask);
        mask+= mask;
    }
}

/*** Subclasses Menu ***/

void HandleSubclassMenu (w, client_data, call_data)
    Widget w;
    caddr_t client_data;
    XmRowColumnCallbackStruct *call_data;
{
    int        chosen_entry = ButtonIndex (buttons6, call_data);
    ClassDoc  *doc = ActiveDoc ();

    ChangeClassChain (doc->subs [chosen_entry]);
}

/* Build a new subclasses menu for the current class */

void SetSubclassesMenu (doc)
    ClassDoc  *doc;
{
    int        i;
    Arg        args[NUM_ARGS];

    XtUnmanageChildren (buttons6, MAX_MENU_BUTTONS);
    for (i=0; i<doc->no_subs; i++) {
        XtSetArg (args[0], XmNlabelString, doc->subs[i]->xm_name);
        XtSetValues (buttons6[i], args, 1);
    }
    XtManageChildren (buttons6, doc->no_subs); 
} 

/*** Help Menu ***/

void HandleHelpMenu (w, client_data, call_data)
    Widget w;
    caddr_t client_data;
    XmRowColumnCallbackStruct *call_data;
{
    Arg    args[NUM_ARGS];
    String buffer;

    buffer = ReadLongSection ("help", "Version");
    PopupHelpDialog (buffer);
}

/************************* Menu Creation ********************************/

/* labels for the menus */

#define FILE_MENU       "File"
#define EDIT_MENU       "Edit"
#define LOCATE_MENU     "Locate"
#define OPTIONS_MENU    "Options"
#define INFO_MENU       "Info"
#define SUBCLASSES_MENU "Subclasses"
#define HELP_MENU       "Help"

static String  menu1[] = 
    {"New", "Open...", "Save", "Save As...", "Print...", "Exit"};

static String  menu2[] = 
    {"Undo", "Cut", "Copy", "Paste", "", "Clear", "Delete"};

static String menu3[] = 
    {"Widget Class...", "Resource Name..."};

static String menu4[] = 
    {"Class", "Type", "Default", "Access", "Range"};

static String menu5[] =
    {"Description", "Behaviour", "Callback", "Translations", "Traversal",
     "Color"};

static String menu6[] = /* only placeholders */
    {"1", "2", "3", "4", "5", "6", "7", "8"};

static String menu7[] =
    {"On Context", "On Window", "On Keys", "Index", "On Help", "Tutorial",
     "On Version"};

/* determine default mnemonic */

#define BLANK ' '
#define NULLC '\0'

char DefaultMnemonic (label)
    String    label;
{
    char     *pos = label;

    while (*pos != NULLC) 
        if (*pos++ == BLANK) return *pos;
    return *label;
}

/* create one pulldown menu */

Widget CreateOneMenu (parent, name, labels, num_labels, buttons, is_toggle)
    Widget      parent;
    char       *name;
    char       *labels[];
    int         num_labels;
    Widget      buttons[10];
    Boolean     is_toggle;

{
    Widget      shell, casc;
    Arg         args[NUM_ARGS];
    int         row;
    XmString    buffer;

    /* create pulldown menu */
    shell = XmCreatePulldownMenu (parent, name, NULL, 0);

    /* create buttons */
    for (row = 0; row < num_labels; row++) {
      if (strlen (labels[row]) == 0) {
        buttons[row] = XmCreateSeparatorGadget (shell, "msep", NULL, 0); }
      else {	  
        buffer = MakeXmString (labels[row]);
        XtSetArg (args[0], XmNlabelString, buffer);
        XtSetArg (args[1], XmNmnemonic, DefaultMnemonic (labels[row]));
        if (is_toggle) {
          XtSetArg (args[2], XmNset, True);
          buttons[row] = XmCreateToggleButtonGadget (shell, "mtoggle",
                                                     args, 3);}
        else {
          buttons[row] = XmCreatePushButtonGadget (shell, "mbutton",
                                                   args, 2);
        }
        XmStringFree (buffer);
      }
    }
    /* manage all buttons */
    XtManageChildren (buttons, num_labels);

    /* create cascade button */
    buffer = MakeXmString (name);
    XtSetArg (args[0], XmNlabelString, buffer);
    XtSetArg (args[1], XmNmnemonic, *name);
    XtSetArg (args[2], XmNsubMenuId, shell);
    casc = XtCreateManagedWidget ("casbutton", 
                xmCascadeButtonWidgetClass, parent, args, 3);
    XmStringFree (buffer);

    if (strcmp (name, HELP_MENU) == 0) {
        XtSetArg (args[0], XmNmenuHelpWidget, casc);
        XtSetValues (parent, args, 1);
    }

    return shell;
}

/* set accelerator for button */

void SetAccelerator (button, acc, acc_text)
    Widget    button;
    char     *acc;
    char     *acc_text;
{
    Arg       args[NUM_ARGS];
    XmString  s;

    s = MakeXmString (acc_text);
    XtSetArg (args[0], XmNaccelerator, acc);
    XtSetArg (args[1], XmNacceleratorText, s);
    XtSetValues (button, args, 2);
    XmStringFree (s);
}

/* create all menus */

Widget CreateMenus (parent)
    Widget      parent;
{
    Widget      bar, menu;
    int         i;
    Arg         args[NUM_ARGS];

    bar = XmCreateMenuBar (parent, "bar", NULL, 0);
    XtManageChild (bar);

    /* create file menu */
    menu = CreateOneMenu (bar, FILE_MENU, 
                          menu1, XtNumber (menu1), buttons1, False);
    XtAddCallback (menu, XmNentryCallback, HandleFileMenu, NULL);
    XtSetSensitive (buttons1[0], False);
    XtSetSensitive (buttons1[1], False);
    XtSetSensitive (buttons1[2], False);
    XtSetSensitive (buttons1[3], False);
    XtSetSensitive (buttons1[4], False);

    /* create edit menu */
    menu = CreateOneMenu (bar, EDIT_MENU, 
                          menu2, XtNumber (menu2), buttons2, False);
    XtAddCallback (menu, XmNentryCallback, HandleEditMenu, NULL);
    XtSetSensitive (buttons2[0], False);
    SetAccelerator (buttons2[0], "Meta<Key>Backspace:", "Alt+Backspace");
    XtSetSensitive (buttons2[1], False);
    SetAccelerator (buttons2[1], "Shift<Key>Delete:", "Shift+Del");
    XtSetArg (args[0], XmNmnemonic, 't');
    XtSetValues (buttons2[1], args, 1);
    XtSetSensitive (buttons2[2], False);
    SetAccelerator (buttons2[2], "Ctrl<Key>Insert:", "Ctrl+Ins");
    XtSetSensitive (buttons2[3], False);
    SetAccelerator (buttons2[3], "Shift<Key>Insert:", "Shift+Ins");
    XtSetSensitive (buttons2[5], False);
    XtSetArg (args[0], XmNmnemonic, 'e');
    XtSetValues (buttons2[5], args, 1);
    XtSetSensitive (buttons2[6], False);

    /* create locate menu */
    menu = CreateOneMenu (bar, LOCATE_MENU, 
                          menu3, XtNumber (menu3), buttons3, False);
    XtAddCallback (menu, XmNentryCallback, HandleLocateMenu, NULL);
    SetAccelerator (buttons3[0], "<Key>F5:", "F5");
    SetAccelerator (buttons3[1], "<Key>F6:", "F6");

    /* create options menu */
    menu = CreateOneMenu (bar, OPTIONS_MENU,
                          menu4, XtNumber (menu4), buttons4, True);
    XtAddCallback (menu, XmNentryCallback, HandleOptionMenu, NULL);

    /* create info menu */
    menu = CreateOneMenu (bar, INFO_MENU,
                          menu5, XtNumber (menu5), buttons5, False);
    XtAddCallback (menu, XmNentryCallback, HandleInfoMenu, NULL);
    XtSetArg (args[0], XmNmnemonic, 'l');
    XtSetValues (buttons5[3], args, 1);

    /* create subclasses menu */
    menu = CreateOneMenu (bar, SUBCLASSES_MENU, 
                          menu6, XtNumber (menu6), buttons6, False);
    XtAddCallback (menu, XmNentryCallback, HandleSubclassMenu, NULL);

    /* create help menu */
    menu = CreateOneMenu (bar, HELP_MENU,
                          menu7, XtNumber (menu7), buttons7, False);
    XtAddCallback (menu, XmNentryCallback, HandleHelpMenu, NULL);
    XtSetSensitive (buttons7[0], False);
    XtSetSensitive (buttons7[1], False);
    XtSetSensitive (buttons7[2], False);
    XtSetSensitive (buttons7[3], False);
    XtSetSensitive (buttons7[4], False);
    XtSetSensitive (buttons7[5], False);

    return bar;
}

/************************** Main Form **************************************/

#define LABEL_CLASS_CHAIN "Superclass Chain"
#define LABEL_RES_FIELDS  "Resource Fields"
#define LABEL_EMPTY_NAME  "XmN..."

Widget param_label;

void CreateMainForm (parent)
    Widget   parent;
{
    Arg      args[NUM_ARGS];
    XmString s;
    int      n;
    Widget   form;
    Widget   class_label, res_label;
    Widget   class_row, res_row, param_row;
    Widget   class_frame, param_frame;

    form = XtCreateManagedWidget ("form", xmFormWidgetClass, parent,
                                  args, 0);

    /* create XmString for label */
    s = MakeXmString (LABEL_CLASS_CHAIN);
    /* setup arguments for label widget */
    n=0;
    XtSetArg (args[n], XmNlabelString, s); n++;
    XtSetArg (args[n], XmNresizable, False); n++;
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_NONE); n++;
    XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++; 
    class_label = XtCreateManagedWidget ("listLabel", 
        xmLabelWidgetClass, form, args, n);
    XmStringFree (s);

    n=0;
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNtopWidget, class_label); n++;
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg (args[n], XmNshadowType, XmSHADOW_IN); n++;
    class_frame = XtCreateManagedWidget ("classFrame", xmFrameWidgetClass, 
                                         form, args, n);
    n=0;
    XtSetArg (args[n], XmNorientation, XmVERTICAL); n++;
    XtSetArg (args[n], XmNspacing, 0); n++;
    XtSetArg (args[n], XmNradioBehavior, True); n++;
    XtSetArg (args[n], XmNradioAlwaysOne, True); n++;
    XtSetArg (args[n], XmNentryAlignment, XmALIGNMENT_CENTER); n++;
    class_row = XtCreateManagedWidget ("classRow", xmRowColumnWidgetClass,
                                       class_frame, args, n);
    CreateClassChain (class_row);
    XmAddTabGroup (class_row);

    /* create status pane ... */
    n=0;
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNtopWidget, class_label); n++;
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg (args[n], XmNshadowType, XmSHADOW_IN); n++;
    param_frame = XtCreateManagedWidget ("pFrame", xmFrameWidgetClass, 
                                         form,  args, n);

    n=0;
    param_row = XtCreateManagedWidget ("pRow", xmRowColumnWidgetClass,
                                       param_frame, args, n);

    /* create XmString for label */
    s = MakeXmString (LABEL_EMPTY_NAME);
    /* setup arguments for label widget */
    n=0;
    XtSetArg (args[n], XmNlabelString, s); n++;
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_NONE); n++;
    XtSetArg (args[n], XmNleftAttachment, XmATTACH_OPPOSITE_WIDGET); n++; 
    XtSetArg (args[n], XmNleftWidget, param_frame); n++;
    XtSetArg (args[n], XmNleftOffset, 0); n++;
    param_label = XtCreateManagedWidget ("pLabel", 
        xmLabelWidgetClass, form, args, n);
    XmStringFree (s);

    CreateParameterField (param_row);

    /* create XmString for label */
    s = MakeXmString (LABEL_RES_FIELDS);
    /* setup arguments for label widget */
    n=0;
    XtSetArg (args[n], XmNlabelString, s); n++;
    XtSetArg (args[n], XmNresizable, False); n++;
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_NONE); n++;
    XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++; 
    XtSetArg (args[n], XmNleftWidget, class_frame); n++;
    res_label = XtCreateManagedWidget ("listLabel", 
        xmLabelWidgetClass, form, args, n);
    XmStringFree (s);

    /* setup arguments for scrolled window */
    n=0;
    XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNtopWidget, res_label); n++;
    XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg (args[n], XmNrightAttachment, XmATTACH_WIDGET); n++;
    XtSetArg (args[n], XmNrightWidget, param_frame); n++;
    XtSetArg (args[n], XmNleftAttachment, XmATTACH_WIDGET); n++; 
    XtSetArg (args[n], XmNleftWidget, class_frame); n++;
    /* create scrolled window */
    res_row = XtCreateManagedWidget ("resSW", xmScrolledWindowWidgetClass,
                                     form, args, n);

    CreateSelectionList (res_row);
}

void SetParamLabel (doc, entry)
    ClassDoc    *doc;
    int          entry;
{
    Arg          args[NUM_ARGS];
    XmString     s;
    char         strbuf[SMALL_BUFLENGTH];

    strcpy (strbuf, "XmN");
    strcat (strbuf, doc->res_name[entry]);
    s = MakeXmString (strbuf);
    XtSetArg (args[0], XmNlabelString, s);
    XtSetValues (param_label, args, 1);
    XmStringFree (s);
}    

/************************* Superclass Chain  *****************************/

/* superclass chain parameters */

#define MAX_CLASS_DEPTH 11
#define MAX_CLASS_WIDGETS 2*MAX_CLASS_DEPTH-1

/* alternating buttons and separators */

static Widget class_chain[MAX_CLASS_WIDGETS];
static ClassDoc *cl_docs[MAX_CLASS_WIDGETS];

/* the activated (pressed) class or -1 */

static int active_class;

/* forward declarations for this section */

void ArmClass ();
void ActivateClass ();

/* report the active class node */

ClassDoc *ActiveDoc ()
{
    if (active_class >= 0)
        return cl_docs[active_class];
    else
        XtErrorMsg (
            NErrInternal, TErrChain,   /* resource name         */
            ERROR_CLASS,               /* resource class        */
            "no active class",         /* default message       */
            NULL, NULL);               /* parameters            */
}

/* Callback for inverting foreground and background */

void InvertToggle (w, client_data,  call_data)
    Widget        w;
    caddr_t       client_data;
    XmToggleButtonCallbackStruct  *call_data;
{
    Arg           args[NUM_ARGS];
    Pixel         fg, bg;

    XtSetArg (args[0], XmNforeground, &fg);
    XtSetArg (args[1], XmNbackground, &bg);
    XtGetValues (w, args, 2);
    XtSetArg (args[0], XmNforeground, bg);
    XtSetArg (args[1], XmNbackground, fg);
    XtSetValues (w, args, 2);
}

/* arm callback procedure for drawn button */

void ArmClass (w, client_data,  call_data)
    Widget        w;
    caddr_t       client_data;
    caddr_t       call_data;
{
    Arg           args[NUM_ARGS];
    int           index = (int) client_data;

    ActivateClass (index);
}

/* create widgets for class chain */

void CreateClassChain (parent)
    Widget         parent;
{
    int            i;
    Arg            args[NUM_ARGS];

    for (i=0; i<MAX_CLASS_DEPTH; i++) {
      XtSetArg (args[0], XmNindicatorOn, False);
      XtSetArg (args[1], XmNshadowThickness, 2);
      XtSetArg (args[2], XmNfillOnSelect, True);
      class_chain[2*i] =
        XtCreateWidget ("chain", xmToggleButtonWidgetClass, parent, args, 3);
      XtAddCallback (class_chain[2*i], XmNarmCallback, ArmClass, 2*i);
      XtAddCallback (class_chain[2*i], XmNvalueChangedCallback, 
                                       InvertToggle, NULL);
      XtSetArg (args[0], XmNorientation, XmVERTICAL);
      class_chain [2*i+1] =
        XtCreateWidget ("c_sep", xmSeparatorGadgetClass, parent, args, 1);
    }
    active_class = -1;
}

/* make one widget the active class */

void ActivateClass (chain_pos)
    int           chain_pos;
{
    ClassDoc     *doc;

    active_class = chain_pos;
    /* read help if necessary */
    doc = cl_docs[active_class];
    FillHelpInfo (doc);
    /* set the selection list contents */
    SetResourceSelectionList (doc);
    /* sensitivity in info menu */
    SetSectionSensitivity (doc);
    /* create subclasses menu */
    SetSubclassesMenu (doc);
    /* clear help text */
    SetHelpText ("");
}

/* build new superclass chain to a newly selected widget class */

void ChangeClassChain (class_doc)
    ClassDoc  *class_doc;
{
    int        i, level;
    Boolean    is_set;
    Arg        args[NUM_ARGS];
    Widget     parent_row = XtParent (class_chain[0]);

    /* inhibit resize */
    if (XtIsRealized (parent_row)) {
        XtSetArg (args[0], XmNresizeWidth, False);
        XtSetValues (parent_row, args, 1);
    }
    /* first unmanage all */
    XtUnmanageChildren (&class_chain[1], MAX_CLASS_WIDGETS-1);
    if (active_class >= 0) 
        InvertToggle (class_chain[active_class], NULL, NULL); 

    /* set button labels downwards from level */
    level = class_doc->level;
    for (i=level-1; i>=0; i--) {
        XtSetArg (args[0], XmNlabelString, class_doc->xm_name);
        is_set = (level == i+1 ? True : False);
        XtSetArg (args[1], XmNset, is_set);
        XtSetValues (class_chain [2*i], args, 2);
        cl_docs[2*i] = class_doc;
        class_doc = class_doc->super;
    }
    /* manage rest and activate last one */
    if (level > 0) {
        XtManageChildren (class_chain, 2*level-1);
        InvertToggle (class_chain[2*level-2], NULL, NULL);
        ActivateClass (2*level-2);
    }
}

/*********************** Resource Selection List ***************************/

/* widget for the selection */

static Widget res_list;

/* callback for list selection */

void GetResourceHelp ();

/* create the selection list */

void CreateSelectionList (parent)
    Widget         parent;
{
    int            n;
    XmString       s;
    Arg            args[NUM_ARGS];
    Widget         s_bar;

    /* create list widget */
    XtSetArg (args[0], XmNscrollBarDisplayPolicy, XmSTATIC);
    XtSetArg (args[1], XmNselectionPolicy, XmSINGLE_SELECT);
    res_list = XtCreateManagedWidget ("list", xmListWidgetClass, parent,
                                       args, 2);
    XmAddTabGroup (res_list);
    XtAddCallback (res_list, XmNsingleSelectionCallback,
                   GetResourceHelp, NULL);

    XtSetArg (args[0], XmNverticalScrollBar, &s_bar);
    XtGetValues (res_list, args, 1);
    if (s_bar != NULL) XmAddTabGroup (s_bar);
}

/* select a resource entry by program */

void SelectResourceEntry (doc, entry)
    ClassDoc    *doc;
    int          entry;
{
    Arg          args[NUM_ARGS];

    XtSetArg (args[0], XmNselectedItems, &doc->res_string[entry]);
    XtSetArg (args[1], XmNselectedItemCount, 1);
    XtSetValues (res_list, args, 2);
    SetHelpText (doc->res_text[entry]);
    SetParameterField (doc, entry);
}

/* Set the list of resource names */

void SetResourceSelectionList (doc)
    ClassDoc      *doc;
{
    Arg            args[NUM_ARGS];

    XtSetArg (args[0], XmNitems, doc->res_string);
    XtSetArg (args[1], XmNitemCount, doc->no_res);
    XtSetValues (res_list, args, 2);
}

/* Display corresponding help text and parameters */

void GetResourceHelp (w, client_data, call_data)
    Widget                w;
    caddr_t               client_data;
    XmListCallbackStruct *call_data;
{
    ClassDoc   *doc = ActiveDoc ();

    SetHelpText (doc->res_text[call_data->item_position - 1]);
    SetParameterField (doc, call_data->item_position - 1);
}

/************************** Parameter Field ***************************/

#define NUM_CAT 5
#define NUM_CHOICES 3

static String text_fields[] =
    {"Class:   ", "Type:    ", "Default: ", "Access:  ", "Range:   "};
static String access_enum[] = {"Create", "Set", "Get"};
static String range_enum[] = {"Constraint", "Scrolled", "Special"};

static Widget prompt_w[NUM_CAT], value_w[NUM_CAT];
static Widget access_choices[NUM_CHOICES], range_choices[NUM_CHOICES];

#define MIDDLE_POS 60

/* create parameter fields */

void CreateParameterField (parent)
    Widget       parent;
{
    int          i, j, n;
    XmString     s;
    Arg          args[NUM_ARGS];

    for (i=0; i<NUM_CAT; i++) {
        /* create label part */
        s = MakeXmString (text_fields[i]);
        n=0;
        XtSetArg (args[n], XmNlabelString, s); n++;
        XtSetArg (args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
        prompt_w[i] = XtCreateWidget ("prompt", xmLabelGadgetClass,
                         parent, args, n);
        XmStringFree(s);

        /* create value part */
        if (i==NUM_CAT-2) {
          /* create first choice */
          n=0;
          XtSetArg (args[n], XmNpacking, XmPACK_COLUMN); n++;
          XtSetArg (args[n], XmNorientation, XmHORIZONTAL); n++;
          XtSetArg (args[n], XmNentryAlignment, XmALIGNMENT_CENTER); n++;
          value_w[i] = XtCreateWidget ("access", xmRowColumnWidgetClass,
                                       parent, args, n);
          for (j=0; j<NUM_CHOICES; j++) {
            s = MakeXmString (access_enum[j]);
            XtSetArg (args[0], XmNlabelString, s);
            access_choices[j] = XtCreateWidget ("choice", 
                                         xmToggleButtonGadgetClass,
                                         value_w[i], args, 1);
            XmStringFree (s);
          }
          XtManageChildren (access_choices, NUM_CHOICES);}
        else if (i==NUM_CAT-1) {
          n=0;
          XtSetArg (args[n], XmNpacking, XmPACK_COLUMN); n++;
          XtSetArg (args[n], XmNorientation, XmHORIZONTAL); n++;
          XtSetArg (args[n], XmNentryAlignment, XmALIGNMENT_CENTER); n++;
          value_w[i] = XtCreateWidget ("range", xmRowColumnWidgetClass,
                                       parent, args, n);
          for (j=0; j<NUM_CHOICES; j++) {
            s = MakeXmString (range_enum[j]);
            XtSetArg (args[0], XmNlabelString, s);
            range_choices[j] = XtCreateWidget ("choice", 
                                         xmDrawnButtonWidgetClass,
                                         value_w[i], args, 1);
            XmStringFree (s);
          }
          XtManageChildren (range_choices, NUM_CHOICES);}
        else {
          /* create text field */
          s = MakeXmString (" ");
          n=0;
          XtSetArg (args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
          XtSetArg (args[n], XmNshadowThickness, 2); n++;
          XtSetArg (args[n], XmNlabelString, s); n++; 
          value_w [i] = XtCreateWidget ("value", xmDrawnButtonWidgetClass,
                                        parent, args, n);
          XmStringFree (s);
        }
    }
    XtManageChildren (value_w, NUM_CAT);
    XtManageChildren (prompt_w, NUM_CAT); 

}

/* fill in parameters according to info in doc node */

void SetParameterField (doc, entry)
    ClassDoc    *doc;
    int          entry;
{
    Arg          args[NUM_ARGS];
    int          i, mask, acc;

    SetParamLabel (doc, entry);

    XtSetArg (args[0], XmNlabelString, doc->res_class[entry]);
    XtSetValues (value_w[0], args, 1);
    XtSetArg (args[0], XmNlabelString, doc->res_type[entry]);
    XtSetValues (value_w[1], args, 1);
    XtSetArg (args[0], XmNlabelString, doc->res_default[entry]);
    XtSetValues (value_w[2], args, 1);
    acc = doc->res_access[entry];
    mask = 1;
    for (i=0; i<NUM_CHOICES; i++) {
        XtSetArg (args[0], XmNset, acc&mask);
        XtSetValues (access_choices[i], args, 1);
        mask+=mask;
    }
    for (i=0; i<NUM_CHOICES; i++) {
        XtSetSensitive (range_choices[i], i == doc->res_range[entry]-1);
    }
}

/* return visibility of one parameter */

Boolean ParamIsVisible (index)
    int        index;
{
    return XtIsManaged (prompt_w [index]);
}

/* change visibility of one parameter */

void ParamVisibility (index, on)
    int        index;
    Boolean    on;
{
    if (on) {
        XtManageChild (prompt_w[index]);
        XtManageChild (value_w[index]);}
    else {
        XtUnmanageChild (prompt_w[index]);
        XtUnmanageChild (value_w[index]);}
}

/******************************** Help Text *******************************/

static Widget helptext;

void CreateHelpText (parent)
    Widget       parent;
{
    Arg          args[NUM_ARGS];
    Widget       h_sb, v_sb;

    XtSetArg (args[0], XmNeditable, False);
    XtSetArg (args[1], XmNeditMode, XmMULTI_LINE_EDIT);
    helptext = XmCreateScrolledText (parent, "helptext", args, 2);
    XtManageChild (helptext);

    XtSetArg (args[0], XmNverticalScrollBar, &v_sb);
    XtSetArg (args[1], XmNhorizontalScrollBar, &h_sb);
    XtGetValues (XtParent (helptext), args, 2);
    if (v_sb != NULL) XmAddTabGroup (v_sb);
    if (h_sb != NULL) XmAddTabGroup (h_sb);
}

void SetHelpText (text)
    String       text;
{
    Arg          args[NUM_ARGS];

    XtSetArg (args[0], XmNvalue, text);
    XtSetValues (helptext, args, 1);
}

/************************ Class Selection Dialog ***************************/

#define LIST_LABEL "Motif Classes"
#define LIST_TITLE "Widget Class Selection"

static Widget class_selection_dialog;

void SelectClass ();


/* create class selection dialog */

void CreateClassSelectionDialog (parent)
    Widget     parent;
{
    Arg        args[NUM_ARGS];
    XmString   s1, s2;
    Widget     w;
    XmString  *class_names;
    int        no_class_names;

    s1 = MakeXmString (LIST_LABEL);
    s2 = MakeXmString (LIST_TITLE);
    class_names = GetClassNames (&no_class_names);
    XtSetArg (args[0], XmNlistItems, class_names);
    XtSetArg (args[1], XmNlistItemCount, no_class_names);
    XtSetArg (args[2], XmNlistLabelString, s1);
    XtSetArg (args[3], XmNdialogTitle, s2);
    class_selection_dialog = XmCreateSelectionDialog (parent, "seldia",
        args, 4);
    XmStringFree (s1);
    XmStringFree (s2);
    XtAddCallback (class_selection_dialog, XmNokCallback, SelectClass, NULL);

    /* manually remove unwanted children */
    w = 
    XmSelectionBoxGetChild (class_selection_dialog, XmDIALOG_SELECTION_LABEL);
    XtUnmanageChild (w);
    w = XmSelectionBoxGetChild (class_selection_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild (w);
    w = XmSelectionBoxGetChild (class_selection_dialog, XmDIALOG_TEXT);
    XtUnmanageChild (w);
}

/* reset class dialogue before popup */

void PopupClassDialog ()
{
    XtManageChild (class_selection_dialog);
}

/* select a new class as chain endpoint */

void SelectClass (w, client_data, call_data)
    Widget     w;
    caddr_t    client_data;
    XmSelectionBoxCallbackStruct *call_data;
{
    ClassDoc  *doc = SearchClass (call_data->value);
    if (doc == NULL) return;
    ChangeClassChain (doc);
}

/*********************** Resource Selection Dialog ************************/

#define RESL_LABEL  "Matching Resource Fields"
#define WORK_LABEL  "Enter part of Resource Name:"

#define RES_TITLE   "Resource Selection"
#define APPLY_LABEL "Search"

static Widget res_selection_dialog;

/* callback for selecting one of the matching resources */

static void SelectRes (w, client_data, call_data)
    Widget     w;
    caddr_t    client_data;
    XmSelectionBoxCallbackStruct   *call_data;
{
    ClassDoc  *doc;
    int        entry;

    if (call_data->value != NULL) {
        LocateResource (call_data->value, &doc, &entry);
        if (doc != NULL) {
            XtUnmanageChild (w);
            ChangeClassChain (doc);
            SelectResourceEntry (doc, entry);
	}
    }
}

/* callback for manual popdown of dialogue */

static void DiaDown (w, client_data, call_data)
    Widget     w;
    caddr_t    client_data;
    caddr_t    call_data;
{
    XtUnmanageChild (w);
}

/* callback for creating list of matching resources */

static void UpdateResList (w, client_data, call_data)
    Widget     w;
    caddr_t    client_data;
    XmSelectionBoxCallbackStruct   *call_data;
{
    Arg        args[NUM_ARGS];
    Widget     wid;
    String     seg;
    XmString  *match_res;
    int        no_match;

    XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET, &seg);
    FindMatchingFields (seg, &match_res, &no_match);
    XtSetArg (args[0], XmNlistItems, match_res); 
    XtSetArg (args[1], XmNlistItemCount, no_match);
    XtSetValues (res_selection_dialog, args, 2);

    if (no_match > 0) {
        wid = XmSelectionBoxGetChild (w, XmDIALOG_OK_BUTTON);
        XtSetArg (args[0], XmNdefaultButton, wid);
        XtSetValues (w, args, 1);}
    else {
        wid = XmSelectionBoxGetChild (w, XmDIALOG_APPLY_BUTTON);
        XtSetArg (args[0], XmNdefaultButton, wid);
        XtSetValues (w, args, 1);
    }
}

/* create the resource selection dialog */

void CreateResSelectionDialog (parent)
    Widget     parent;
{
    Arg        args[NUM_ARGS];
    XmString   s1, s2, s3, s4;
    Widget     w;

    s1 = MakeXmString (RESL_LABEL);
    s2 = MakeXmString (WORK_LABEL);
    s3 = MakeXmString (RES_TITLE);
    s4 = MakeXmString (APPLY_LABEL);
    XtSetArg (args[0], XmNlistLabelString, s1);
    XtSetArg (args[1], XmNautoUnmanage, False);
    XtSetArg (args[2], XmNselectionLabelString, s2);
    XtSetArg (args[3], XmNdialogTitle, s3);
    XtSetArg (args[4], XmNapplyLabelString, s4);
    res_selection_dialog = XmCreateSelectionDialog (parent, "resdia",
        args, 5);
    XmStringFree (s1); XmStringFree (s2);
    XmStringFree (s3); XmStringFree (s4);
    XtAddCallback (res_selection_dialog, XmNokCallback, SelectRes, NULL);
    XtAddCallback (res_selection_dialog, XmNapplyCallback, 
                   UpdateResList, NULL);
    XtAddCallback (res_selection_dialog, XmNcancelCallback, DiaDown, NULL);
    w = XmSelectionBoxGetChild (res_selection_dialog, XmDIALOG_APPLY_BUTTON);
    XtManageChild (w);
    w = XmSelectionBoxGetChild (res_selection_dialog, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild (w);
    XtSetArg (args[0], XmNdefaultButton, w);
    XtSetValues (res_selection_dialog, args, 1);
}

/* reset resource dialog before popup */

void PopupResDialog ()
{
    Widget     w;
    Arg        args[NUM_ARGS];
    XmString   s;

    w = XmSelectionBoxGetChild (res_selection_dialog, XmDIALOG_APPLY_BUTTON);
    s = MakeXmString ("");
    XtSetArg (args[0], XmNdefaultButton, w);
    XtSetArg (args[1], XmNlistItems, NULL);
    XtSetArg (args[2], XmNlistItemCount, 0);
    XtSetArg (args[3], XmNtextString, s);
    XtSetValues (res_selection_dialog, args, 4);
    XmStringFree (s);
    XtManageChild (res_selection_dialog);
}

/***************************** Help Dialog ********************************/

#define HELP_TITLE "Help Text"

static Widget help_dialog;

/* create held dialog */

void CreateHelpDialog (parent)
    Widget     parent;
{
    Arg        args[NUM_ARGS];
    XmString   s1;

    s1 = MakeXmString (HELP_TITLE);
    XtSetArg (args[0], XmNdialogTitle, s1);

    help_dialog = XmCreateMessageDialog (parent, "helpdia", args, 1);
    XmStringFree (s1);
}

/* pop up the help dialog */

void PopupHelpDialog (buffer)
    String     buffer;
{
    Arg        args[NUM_ARGS];
    XmString   s;

    s = MakeXmString (buffer);
    XtSetArg (args[0], XmNmessageString, s);
    XtSetValues (help_dialog, args, 1);
    XmStringFree (s);
    XtManageChild (help_dialog);
}




